import os
from concurrent.futures import ThreadPoolExecutor, as_completed

from newspy import newsorg, rss
from newspy.shared.models import Source, Channel, Article

default_client_config = {}

channels = {
    Channel.NEWSORG: newsorg,
    Channel.RSS: rss,
}


def configure(newsorg_api_key: str | None = None) -> None:
    global default_client_config

    if newsorg_api_key is None:
        newsorg_api_key = os.getenv("NEWSORG_API_KEY")

    default_client_config = {
        "newsorg_api_key": newsorg_api_key,
    }


def get_sources() -> list[Source]:
    sources = []
    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(channels[key].client.get_sources) for key in channels
        ]

        for future in as_completed(futures):
            sources.extend([r.to_source() for r in future.result()])

    return sources


def get_articles() -> list[Article]:
    articles = []
    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(channels[key].client.get_articles) for key in channels
        ]

        for future in as_completed(futures):
            articles.extend([r.to_article() for r in future.result()])

    return articles



import csv
import gzip
import io
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import NewType

from newspy.rss.models import RssSource, RssArticle
from newspy.shared.http_client import HttpClient, HttpMethod

URL = NewType("URL", str)


def get_articles(sources: list[RssSource] | None = None) -> list[RssArticle]:
    if not sources:
        sources = get_sources()

    http_client = HttpClient()

    articles = []
    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(
                http_client.send,
                method=HttpMethod.GET,
                url=source.url,
                headers={"Content-Type": "application/rss+xml"},
                params=None,
                payload=None,
            )
            for source in sources
        ]

        for future in as_completed(futures):
            resp_json = future.result()

            if resp_json:
                for article in resp_json:
                    try:
                        source = filter(
                            lambda s: s.url == article["source_url"], sources
                        )
                        articles.append(
                            RssArticle(
                                source=next(source),
                                title=article["title"],
                                description=article["description"],
                                url=article["url"],
                                published=article["published"],
                            )
                        )
                    except KeyError:
                        pass

    return articles


def get_sources(
    file_path: (Path | URL) = URL(
        "https://github.com/onemoola/newspy/blob/main/data/rss_sources.csv.gz?raw=true"
    ),
) -> list[RssSource]:
    if isinstance(file_path, Path):
        file_content = file_path
    elif isinstance(file_path, str):
        http_client = HttpClient()
        file_content = http_client.send(
            method=HttpMethod.GET,
            url=file_path,
            headers={"Content-Type": "application/zip"},
        )
        file_content = io.BytesIO(file_content)
    else:
        return []

    with gzip.open(file_content, "rt") as f:
        reader = csv.DictReader(f)
        sources = [RssSource(**row) for row in reader]

    return sources



from dataclasses import dataclass
from datetime import datetime
from enum import Enum


class Language(str, Enum):
    AR = "ar"
    DE = "de"
    EN = "en"
    ES = "es"
    FR = "fr"
    HE = "he"
    IT = "it"
    NL = "nl"
    NO = "no"
    PT = "pt"
    RU = "ru"
    SV = "sv"
    UD = "ud"
    ZH = "zh"


class Country(str, Enum):
    AE = "ae"
    AR = "ar"
    AT = "at"
    AU = "au"
    BE = "be"
    BG = "bg"
    BR = "br"
    CA = "ca"
    CH = "ch"
    CO = "co"
    CU = "cu"
    CZ = "cz"
    DE = "de"
    EG = "eg"
    FR = "fr"
    GB = "gb"
    GR = "gr"
    HK = "hk"
    HU = "hu"
    ID = "id"
    IE = "ie"
    IL = "il"
    IN = "in"
    IT = "it"
    JP = "jp"
    KR = "kr"
    LT = "lt"
    LV = "lv"
    MA = "ma"
    MX = "mx"
    MY = "my"
    NG = "ng"
    NL = "nl"
    NO = "no"
    NZ = "nz"
    PH = "ph"
    PL = "pl"
    PT = "pt"
    RO = "ro"
    RS = "rs"
    RU = "ru"
    SA = "sa"
    SE = "se"
    SG = "sg"
    SI = "si"
    SK = "sk"
    TH = "th"
    TR = "tr"
    TW = "tw"
    UA = "ua"
    US = "us"
    VE = "ve"
    ZA = "za"


class Category(str, Enum):
    BUSINESS = "business"
    FINANCIAL = "financial"
    ENTERTAINMENT = "entertainment"
    GENERAL = "general"
    HEALTH = "health"
    SCIENCE = "science"
    SPORTS = "sports"
    TECHNOLOGY = "technology"


class Channel(str, Enum):
    """Source domain model"""

    NEWSORG = "NEWSORG"
    RSS = "RSS"


@dataclass
class Source:
    """Publisher domain model"""

    id: str | None
    name: str
    channel: Channel


@dataclass
class Article:
    """Publication domain model"""

    slug: str
    url: str
    url_to_image: str | None
    title: str
    abstract: str
    author: str | None
    source: Source
    published: datetime



import json
from enum import Enum
from xml.etree import ElementTree

import requests
from requests.adapters import HTTPAdapter, Retry

from newspy.shared.exceptions import NewspyHttpException


class ContentType(str, Enum):
    JSON = "application/json"
    XML = "application/xml"


class HttpMethod(str, Enum):
    GET = "GET"
    POST = "POST"


class HttpClient:
    MAX_RETRIES = 3

    def __init__(
        self,
        requests_session: bool = True,
        requests_timeout: int = 5,
        status_forcelist: tuple = (429, 500, 502, 503, 504),
        retries: int = MAX_RETRIES,
        status_retries: int = MAX_RETRIES,
        backoff_factor: float = 0.3,
    ) -> None:
        """
        :param requests_timeout:
            Tell Requests to stop waiting for a response after a given
            number of seconds
        :param status_forcelist:
            Tell requests what type of status codes retries should occur on
        :param retries:
            Total number of retries to allow
        :param status_retries:
            Number of times to retry on bad status codes
        :param backoff_factor:
            A backoff factor to apply between attempts after the second try
            See urllib3 https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html
        """
        self._requests_timeout = requests_timeout
        self._status_forcelist = status_forcelist
        self._retries = retries
        self._status_retries = status_retries
        self._backoff_factor = backoff_factor

        if requests_session:  # Build a new session.
            self._build_session()
        else:  # Use the Requests API module as a "session".
            self._session = requests.api

    def _build_session(self) -> None:
        self._session = requests.Session()
        retry = Retry(
            total=self._retries,
            connect=1,
            read=False,
            allowed_methods=frozenset(["GET", "POST"]),
            status=self._status_retries,
            backoff_factor=self._backoff_factor,
            status_forcelist=self._status_forcelist,
        )

        adapter = HTTPAdapter(max_retries=retry)
        self._session.mount("https://", adapter)

    def send(
        self,
        method: HttpMethod,
        url: str,
        headers: dict | None = None,
        params: dict | None = None,
        payload: dict | None = None,
    ) -> json:
        args = {}
        if headers is None:
            headers = {"Content-Type": "application/json"}

        if payload:
            if headers["Content-Type"] == "application/json":
                args["data"] = json.dumps(payload)
            else:
                args["data"] = str(payload)

        try:
            response = self._session.request(
                method,
                url,
                headers=headers,
                timeout=self._requests_timeout,
                params=params,
                **args,
            )

            response.raise_for_status()

            match headers["Content-Type"]:
                case "application/json":
                    results = response.json()
                case "application/rss+xml":
                    results = parse_xml(response.content)
                case "application/zip":
                    results = response.content
                case _:
                    results = response.text

        except requests.exceptions.HTTPError as http_error:
            response = http_error.response
            try:
                json_response = response.json()
                error = json_response.get("error", {})
                msg = error.get("message")
                reason = error.get("reason")
            except ValueError:
                msg = response.text or None
                reason = None

            raise NewspyHttpException(
                status_code=response.status_code,
                msg="%s:\n %s" % (response.url, msg),
                reason=reason,
                headers=response.headers,
            )
        except requests.exceptions.RetryError as retry_error:
            request = retry_error.request
            try:
                reason = retry_error.args[0].reason
            except (AttributeError, IndexError):
                reason = None
            raise NewspyHttpException(
                status_code=429,
                msg="%s:\n %s" % (request.path_url, "Max Retries"),
                reason=reason,
            )
        except (TypeError, ValueError):
            results = None

        return results


def parse_xml(data: str) -> list[dict[str, str]] | None:
    try:
        root = ElementTree.fromstring(data)
        source_url = root.find("channel/link").text.strip()
        items = []
        for item in root.findall(".//item"):
            title = item.find("title").text.strip()
            description = item.find("description").text.strip()
            link = item.find("link").text.strip()
            pub_date = item.find("pubDate").text.strip()
            items.append(
                {
                    "source_url": source_url,
                    "title": title,
                    "description": description,
                    "url": link,
                    "published": pub_date,
                }
            )
        return items
    except (ElementTree.ParseError, AttributeError):
        # Handle XML parsing errors
        return None



import logging
from dataclasses import dataclass
from enum import Enum

from newspy.shared import utils
from newspy.shared.models import Article, Source, Language, Channel

logger = logging.getLogger(__name__)


@dataclass
class RssArticleMediaContent:
    height: str
    width: str
    url: str
    medium: str | None


@dataclass
class RssArticleContent:
    base: str
    value: str


class RssCategory(str, Enum):
    BUSINESS = "business"
    FINANCIAL = "financial"
    ENTERTAINMENT = "entertainment"
    GENERAL = "general"
    HEALTH = "health"
    SCIENCE = "science"
    SPORTS = "sports"
    TECHNOLOGY = "technology"


@dataclass
class RssSource:
    id: str
    name: str
    description: str
    url: str
    category: RssCategory
    language: Language

    def to_source(self) -> Source:
        return Source(id=self.id, name=self.name, channel=Channel.RSS)


@dataclass
class RssArticle:
    source: RssSource
    title: str
    description: str
    url: str
    published: str

    def to_article(self) -> Article:
        source = self.source.to_source()

        return Article(
            slug=f"{utils.slugify(source.name)}-{utils.slugify(self.title)}",
            url=self.url,
            url_to_image=None,
            title=self.title,
            abstract=self.description,
            author=None,
            source=source,
            published=utils.to_datetime(self.published),
        )



import logging
from datetime import date

from newspy import client
from newspy.newsorg.models import (
    NewsorgArticlesRes,
    NewsorgCategory,
    NewsorgEndpoint,
    NewsorgSourceRes,
    NewsorgSource,
    NewsorgArticle,
)
from newspy.shared.exceptions import NewspyException
from newspy.shared.http_client import HttpClient, HttpMethod
from newspy.shared.models import Country, Language

logger = logging.getLogger(__name__)

BASE_URL = "https://newsapi.org/v2"


def create_url(endpoint: NewsorgEndpoint) -> str:
    match endpoint:
        case NewsorgEndpoint.EVERYTHING:
            return f"{BASE_URL}/everything"
        case NewsorgEndpoint.TOP_HEADLINES:
            return f"{BASE_URL}/top-headlines"
        case _:
            raise NewspyException(
                msg=f"The endpoint has to be one of the following values: 'EVERYTHING' or 'TOP_HEADLINES' or 'SOURCES'",
            )


def create_params(
    endpoint: NewsorgEndpoint = NewsorgEndpoint.TOP_HEADLINES,
    search_text: str | None = None,
    category: NewsorgCategory | None = None,
    country: Country | None = None,
    language: Language | None = None,
    sources: list[NewsorgSource] | None = None,
    from_date: date | None = None,
    to_date: date | None = None,
    page_size: int | None = None,
    page: int | None = None,
) -> dict[str, str]:
    newsorg_api_key = client.default_client_config.get("newsorg_api_key")

    if newsorg_api_key is None:
        raise NewspyException(
            msg="The Newsorg API key is not configured. Please configure it by calling the configure function.",
        )

    if category and sources:
        raise NewspyException(
            msg="Choose either the category and sources attributes. Not both.",
        )

    if country and sources:
        raise NewspyException(
            msg="Choose either the country and sources attributes. Not both.",
        )

    params = {"apiKey": newsorg_api_key}

    if search_text:
        params["q"] = search_text
    if sources and len(sources) > 0:
        params["sources"] = ",".join([source.id for source in sources])

    if endpoint.TOP_HEADLINES:
        if country:
            params["country"] = country.value

        if category:
            params["category"] = category.value

    if endpoint.EVERYTHING:
        if from_date and to_date:
            if from_date > to_date:
                raise NewspyException(
                    msg="The from date cannot be greater than the to date.",
                )

            params["from"] = from_date.strftime("%Y-%m-%d")
            params["to"] = to_date.strftime("%Y-%m-%d")

        if language:
            params["language"] = language.value

    if not page_size:
        page_size = 100
    params["pageSize"] = page_size

    if not page:
        page = 1
    params["page"] = page

    return params


def get_articles(
    endpoint: NewsorgEndpoint = NewsorgEndpoint.TOP_HEADLINES,
    search_text: str | None = None,
    category: NewsorgCategory | None = None,
    country: Country | None = None,
    language: Language | None = None,
    sources: list[NewsorgSource] | None = None,
    from_date: date | None = None,
    to_date: date | None = None,
    page_size: int | None = None,
    page: int | None = None,
) -> list[NewsorgArticle]:
    params = create_params(
        endpoint=endpoint,
        search_text=search_text,
        category=category,
        country=country,
        language=language,
        sources=sources,
        from_date=from_date,
        to_date=to_date,
        page_size=page_size,
        page=page,
    )

    http_client = HttpClient()
    resp_json = http_client.send(
        method=HttpMethod.GET,
        url=create_url(endpoint=endpoint),
        params=params,
    )

    try:
        article_res = NewsorgArticlesRes(**resp_json)
    except TypeError as exc:
        raise NewspyException(
            msg=f"Failed to validate the News Org articles response json: {resp_json}",
            reason=str(exc),
        )

    return article_res.articles


def get_sources(
    category: NewsorgCategory | None = None,
    language: Language | None = None,
    country: Country | None = None,
) -> list[NewsorgSource]:
    params = create_params(category=category, language=language, country=country)

    http_client = HttpClient()
    resp_json = http_client.send(
        method=HttpMethod.GET,
        url=f"{BASE_URL}/top-headlines/sources",
        params=params,
    )

    try:
        source_res = NewsorgSourceRes(**resp_json)
    except TypeError as exc:
        raise NewspyException(
            msg=f"Failed to validate the News Org sources response json: {resp_json}",
            reason=str(exc),
        )

    return source_res.sources



class NewspyHttpException(Exception):

    def __init__(
            self,
            status_code: int,
            msg: str,
            reason: str | None = None,
            headers: dict | None = None,
    ) -> None:
        self._status_code = status_code
        self._msg = msg
        self._reason = reason
        if headers is None:
            headers = {}
        self.headers = headers

    def __str__(self):
        return f"status code: {self._status_code}, message: {self._msg}, reason: {self._reason}"


class NewspyException(Exception):

    def __init__(
            self,
            msg: str,
            reason: str | None = None
    ) -> None:
        self._msg = msg
        self._reason = reason

    def __str__(self) -> str:
        return f"message: {self._msg}, reason: {self._reason}"
